#!/usr/bin/gawk -f

# Generate functions
function gen(head, amount, model) {

	print "\t\tcase " model ":\n"
	print "\t\t\t" head "("

	for (a = amount - 1; a > -1; a--){

		if (and (model, 2^a))
			{print "farg[" amount - 1 - a "]"}
		else
			{print "arg[" amount - 1 - a "]"}
		if (a > 0)
			{print ","}
	}
	print ");\n"
	print "\t\t\tbreak;\n"
}

# Main program
BEGIN {
	ORS = ""
	print "/* Generated with AWK */\n\n"
	print "#include \"gtk-server.h\"\n"
	print "#include <gtk/gtk.h>\n\n"

	# Generate routines for VOID return functions
	print "void Void_Func(void (*func)(), int amount, int type, void *arg[], float farg[])\n"
	print "{\n"
	print "/* Jump to right amount of arguments */\n"

	print "switch (amount){\n\n"
	for (j = 8; j > -1; j--) {

		print "case " j ": /* GTK API call has " j " arguments */\n"
		print "\tswitch (type){\n"

		for (i = 0; i < 2^j; i++){
			gen("(*func)", j, i)
		}
		print "\t}\n"
		print "\tbreak;\n\n"
	}
	print "default:\n"
	print "\tPrint_Error(1, \"ERROR: Too many arguments!\");\n"
	print "}\n"
	print "}\n\n"
	print "/*-------------------------------------------------------------------------------------------*/\n\n"

	# Generate routines for WIDGET return functions
	print "GtkWidget *Widget_Func(GtkWidget *(*func)(), int amount, int type, void *arg[], float farg[])\n"
	print "{\n"
	print "GtkWidget *widget;\n"
	print "/* Jump to right amount of arguments */\n"

	print "switch (amount){\n\n"
	for (j = 8; j > -1; j--) {

		print "case " j ": /* GTK API call has " j " arguments */\n"
		print "\tswitch (type){\n"

		for (i = 0; i < 2^j; i++){
			gen("widget = (GtkWidget*)(*func)", j, i)
		}
		print "\t}\n"
		print "\tbreak;\n\n"
	}
	print "default:\n"
	print "\tPrint_Error(1, \"ERROR: Too many arguments!\");\n"
	print "}\n"
	print "return widget;\n"
	print "}\n\n"
	print "/*-------------------------------------------------------------------------------------------*/\n\n"

	# Generate routines for STRING return functions
	print "char *String_Func(char *(*func)(), int amount, int type, void *arg[], float farg[])\n"
	print "{\n"
	print "char *tmp;\n"
	print "/* Jump to right amount of arguments */\n"

	print "switch (amount){\n\n"
	for (j = 8; j > -1; j--) {

		print "case " j ": /* GTK API call has " j " arguments */\n"
		print "\tswitch (type){\n"

		for (i = 0; i < 2^j; i++){
			gen("tmp = (char*)(*func)", j, i)
		}
		print "\t}\n"
		print "\tbreak;\n\n"
	}
	print "default:\n"
	print "\tPrint_Error(1, \"ERROR: Too many arguments!\");\n"
	print "}\n"
	print "return tmp;\n"
	print "}\n\n"
	print "/*-------------------------------------------------------------------------------------------*/\n\n"

	# Generate routines for BOOLEAN return functions
	print "gboolean Bool_Func(gboolean (*func)(), int amount, int type, void *arg[], float farg[])\n"
	print "{\n"
	print "gboolean boolean;\n"
	print "/* Jump to right amount of arguments */\n"

	print "switch (amount){\n\n"
	for (j = 8; j > -1; j--) {

		print "case " j ": /* GTK API call has " j " arguments */\n"
		print "\tswitch (type){\n"

		for (i = 0; i < 2^j; i++){
			gen("boolean = (*func)", j, i)
		}
		print "\t}\n"
		print "\tbreak;\n\n"
	}
	print "default:\n"
	print "\tPrint_Error(1, \"ERROR: Too many arguments!\");\n"
	print "}\n"
	print "return boolean;\n"
	print "}\n\n"
	print "/*-------------------------------------------------------------------------------------------*/\n\n"

	# Generate routines for LONG return functions
	print "guint Long_Func(guint (*func)(), int amount, int type, void *arg[], float farg[])\n"
	print "{\n"
	print "guint result;\n"
	print "/* Jump to right amount of arguments */\n"

	print "switch (amount){\n\n"
	for (j = 8; j > -1; j--) {

		print "case " j ": /* GTK API call has " j " arguments */\n"
		print "\tswitch (type){\n"

		for (i = 0; i < 2^j; i++){
			gen("result = (*func)", j, i)
		}
		print "\t}\n"
		print "\tbreak;\n\n"
	}
	print "default:\n"
	print "\tPrint_Error(1, \"ERROR: Too many arguments!\");\n"
	print "}\n"
	print "return result;\n"
	print "}\n\n"
	print "/*-------------------------------------------------------------------------------------------*/\n\n"

	# Generate routines for FLOAT return functions
	print "gfloat Float_Func(gfloat (*func)(), int amount, int type, void *arg[], float farg[])\n"
	print "{\n"
	print "gfloat result;\n"
	print "/* Jump to right amount of arguments */\n"

	print "switch (amount){\n\n"
	for (j = 8; j > -1; j--) {

		print "case " j ": /* GTK API call has " j " arguments */\n"
		print "\tswitch (type){\n"

		for (i = 0; i < 2^j; i++){
			gen("result = (*func)", j, i)
		}
		print "\t}\n"
		print "\tbreak;\n\n"
	}
	print "default:\n"
	print "\tPrint_Error(1, \"ERROR: Too many arguments!\");\n"
	print "}\n"
	print "return result;\n"
	print "}\n\n"
	print "/*-------------------------------------------------------------------------------------------*/\n\n"

}
