/**************************************************************************
*
* This is the infamous GTK-SERVER. The purpose of this program is to enable
* access to graphical GTK widgets by using shell scripts. Currently these
* scriptlanguages have been tested successfully:
*
* -KSH (stdin)
* -ZSH (stdin)
* -AWK (stdin, tcp)
* -BASH (tcp, fifo)
* -Cshell (fifo)
* -Rebol (tcp)
* -TCL (tcp)
* -Scriptbasic (tcp, fifo)
* -Prolog (tcp)
* -Python (tcp)
* -newLISP (tcp)
* -CLISP (tcp)
* -PHP (tcp, stdin)
* -VB Script (stdin, fifo)
* -Yabasic (fifo)
* -Perl (fifo)
* -Ash (fifo)
* -Ruby (tcp)
* -Icon (fifo)
* -Logo (fifo)
* -Bas (fifo)
*
* Please read the documentation on how to use this program.
* Many thanks to the following persons for their help: 
* 
* Albert Wessel: pointing to the crucial C library functions
* Norman Deppenbroek: general programming tricks, design optimizations, troubleshooting
* Ronald Meijer: Win32 tips and tricks
* Takafumi Shido: Win32 STDIN trick and CLisp troubleshooting
* Joe Armstrong: major improvements in the callback functionality (1.2)
* James Bailey: writing a flawless s-expression parser (1.3)
*
* Original idea and design by Peter van Eerten, september 2003 - april 2005
* Mail: peter@gtk-server.org
*
* This program is free software; you can redistribute it and/or modify
* it under the terms of the GNU General Public License as published by
* the Free Software Foundation; either version 2 of the License, or
* (at your option) any later version.
*
* This program is distributed in the hope that it will be useful,
* but WITHOUT ANY WARRANTY; without even the implied warranty of
* MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
* GNU General Public License for more details.
*
* You should have received a copy of the GNU General Public License
* along with this program; if not, write to the Free Software
* Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*
* See http://www.gnu.org/licenses/gpl.txt for the full version of this license.
*
* TODO
* - Better errorhandling
*
* CHANGES GTK-SERVER 1.0
* ----------------------
* - BUILD 1:	. Intitial release
* - BUILD 2:	. Fixed bug in parsing GTK-function arguments with Trim_Spaces
* - BUILD 3:	. Removed the child process counter, the Linux TCP version will run standalone now,
*				  and an explicit kill is required.
* - BUILD 4:	. Show help even when configfile is not available
*				. Lines in configfile 'gtk-server.cfg' may be empty now
*				. Name of the external gtk-library is configurable now
* - BUILD 5:	. Changed the format of the gtk-server.cfg file
*				. Added callback signal "RELEASE"
* - BUILD 6:	. Added ability to return a LONG from a GTK-function
*   			. Code cleaning
* - BUILD 7:	. Enable logging facility
* - BUILD 8:	. A lot of code optimizations - source has become smaller
* - BUILD 9:	. Implemented workaround for floating point arguments
*				. Changed logfile name to 'gtk-server.log'
*
* CHANGES GTK-SERVER 1.1
* ----------------------
* - BUILD 1:	. Redesigned the callback mechanism, hence the new release number (1.1)
*				. Updated the Check_Exceptions routine with GTK2.x functions
*				. Changed Widget casting from GtkWidget to GtkObject
*				. Added ability to return a FLOAT from a GTK-function
* - BUILD 2:	. Improved logging facility; returned string will be logged as well
* - BUILD 3:	. Determine maximum clients as an argument to the gtk-server (Linux)
* - BUILD 4:	. Look for global gtk-server.cfg as well (Linux: /etc directory)
* - BUILD 5:	. Path for logfile is configurable now in the gtk-server.cfg file
* - BUILD 6:	. Removed the GTK_LIB_SIGNAL keyword, signals can be configured directly now.
*				. Also search the /usr/local/etc directory for a configfile
* - BUILD 7:	. Added code to compile a self-containing GTK-server with all the
*				  GTK libs included
* - BUILD 8:	. Fixed some illegal memory writes and leaks pointed out by Valgrind
*				. Changed Win32 functionality: server will kill itself if a client
*   			  script disconnects
* - BUILD 9:	. Named pipe support (Unix/Linux)
*				. Some small code optimizations
* - BUILD 10:	. Support for named pipes in Win32
* - BUILD 11:	. Support for STDIN in Win32 (CLisp, Prolog, VB Script)
* - BUILD 12:	. Workaround for passing strings containing comma's as argument - introducing
*   			  the GTK_SERVER_COMMA config option
*   			. Added some GTK2 float functions
*   			. Win32 logic for finding the configfile
* - BUILD 13:	. Improved Named Pipe support for Windows
*   			. Fixed potential bug with creating the logfile
* - BUILD 14:	. Introducing environment variable GTK_SERVER_CONFIG to point to
*   			  configfile (WIN + LINUX)
*				. The Win32 version is not console based anymore. Adjusted all error warnings
*				  to messageboxes.
* - BUILD 15:	. Solved crash when GTK_SERVER_COMMA option is not present
*   			. Installer for Win32
*   			. Autoconf source for Linux
* - BUILD 16:	. Implemented UDP communication interface: gtk-server udp <host:ip> [log]
*				. Argument syntax for TCP has changed: gtk-server tcp <host:ip[:max]> [log]
*				. Improved Makefile for Windows
* - BUILD 17:	. Added float functions for GTK2
*				. Corrected the Linux help prompt
*				. Removed redundant code
* - BUILD 18:	. Centralized version indication of GTK-server
*				. Added internal 'gtk_server_version' command to retrieve version
*				. If a LONG argument contains letters it will switch to STRING automatically
*				. Some large code improvements
*				. Win32: always create a logfile to capture GTK errors
*				. Bugfix: the GTK-server will not crash anymore when sending empty arguments
*				. Bugfix: float values are parsed correctly now in Win32 environments (Linux was already working)
* - BUILD 19:	. Linux: FIFO communication will return result string with newline attached
*				. Linux: improved FIFO to synchronous writing (added O_SYNC flag)
*				. Linux: Added experimental FILE interface (SLOW!)
*				. Win/Linux: Fixed bug with ')' symbol - introducing GTK_SERVER_BRACKET config option
*				. Linux distribution provides convenient KILL script to kill hanging gtk-servers
*
* CHANGES GTK-SERVER 1.2
* ----------------------
* - BUILD 1:	. Major changes in callback functionality by ideas of mr. Joe Armstrong; hence new releasenumber
*				. Ability to add extra signals per widget, each with own identifier
*				. Bugfix: returning empty STRING from GTK widget does not crash GTK-server anymore
*   			. Bugfix: do not crash with '-1' as widget reference
*				. Improvements in the GTK-server KILL-script (Linux)
* - BUILD 2:	. The callback function only returns registered events now if called with 'WAIT'
*   			. The callback function now knows asynchronous "update" as argument
*   			. Generate error message when arguments to GTK-server are not recognized
*   			. Bugfix: reading CONFIG environment variable (Linux)
*				. More improvements in the GTK-server KILL-script (Linux)
*				. HOTFIX: STDIN interface can read newlines now
* - BUILD 3:	. Improved configure script to force compilation with GTK2
*				. Corrected 'gtk-server.cfg' search order for Linux
*				. Corrected format of logging with TCP and UDP
*				. The GTK-server now compiles with -pedantic flag and is ISO-C compatible (Linux)
*				. Added scripts to create sourcepackage, Slackware installpackage and SRC rpm
*   
* CHANGES GTK-SERVER 1.3
* ----------------------
* - BUILD 1:	. Major changes in passing strings to GTK-server; concept by Norman Deppendbroek, parser
*				  created by Jim Bailey, slightly modified for GTK-server (me)
*				. The configoptions 'GTK_SERVER_COMMA' and 'GTK_SERVER_BRACKET' are redundant now
*				. Improved Trim_String function by Jim Bailey
*				. Linux: search for configfile '.gtk-server.cfg' in homedirectory of user also
* - BUILD 2:	. Fixed severe memory leak in callback routines
*				. GTK-server should compile without problems on FreeBSD now
* - BUILD 3:	. Finalized complete solution for FLOAT arguments
*				. Support for opening extra library with GTK_LIB_EXTRA
*				. Removed the experimental FILE interface
* - BUILD 4:	. Fixed memory leak with returning strings (PvE)
*				. Fixed compile warning with GTK1.x (PvE)
*				. Fixed environment variable mixup (Felix Winkelmann, Sebastiaan van Erk)
*				. Fixed STDIN loop exit problems (Sebastiaan van Erk)
*				. Changes in MAN pages (PvE)
*				. Changes in static Makefile (PvE)
*				. Removed FILE interface from GTK-server KILL script (PvE)
*				. Updated gtk-server.cfg file to latest (PvE)
*
* Improvements by others:
* -----------------------
* - 24/01/2004:	Fixed compile problem with GCC 2.96 Redhat 7.1 (thanks Larry Richardson), fix
*   			included since 1.1 build 7.
*   
***************************************************************************/

#include "gtk-server.h"

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <gtk/gtk.h>
#include <locale.h>
#include <unistd.h>
#include <sys/stat.h>

#ifdef LINUX
#include <netdb.h>
#include <sys/socket.h>
#include <sys/wait.h>
#include <arpa/inet.h>
#include <signal.h>
#include <dlfcn.h>
#include <stdarg.h>
/* Needed for FIFO files */
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#endif

#ifdef BSD
#include <netinet/in.h>
#endif

#ifdef WIN32
#include <strings.h>
#include <winsock.h>
#include <windows.h>
#include <shlwapi.h>
#endif

#define STDIN 0

/* Define global constants */
#define MAX_LEN 1024		/* max length of a STDIN coming from script or configfile */

/* Define callback signals */
#define GTK_SERVER_NONE 0

/* Define GTK-server version */
#define VERSION "1.3 build 4"

/* Define backlog for tcp-connections */
#define BACKLOG 4

/* Define current 'callbacked' object */
struct callback {
	GtkWidget *object;
	int state;
	char *text;
	int mousex;
	int mousey;
};

/* Define global instance of current callback */
struct callback Current_Object;

/* Global counter to limit client access */
int *count_fork;

/* Define structure containing configuration data */
typedef struct config_struct {
	int counter;
	char *name;
	char *callbacktype;
	char *returnvalue;
	char *argamount;
	char *args[10];
	struct config_struct *next;
} CONFIG;

/*************************************************************************************************/

/* This function was rewritten by Jim Bailey */

char *Trim_String(char *data)
{
int last_idx;
/* Get rid of whitespaces at the beginning of the string */
while (*data == ' '|| *data == '\t' || *data == '\n') data++;
/* Check there is anything left to chop off at the end */
last_idx = strlen( data ) - 1;
if ( last_idx < 0 ) return data;
/* Find the last non-whitespace at the end of the string */
while ( data[last_idx] == ' ' || data[last_idx] == '\t' || data[last_idx] == '\n' ) last_idx--;
/* Actually chop off the whitespaces at the end of the string */
data[last_idx+1] = '\0';
/* Return stripped string */
return data;
}

/*************************************************************************************************/

/* Check if a charpointer really contains a numbers */

int is_value (char *var) 
{
int i;

if (var != NULL){
	for (i = 0; i < strlen(var); i++){
		if ((*(var + i) < 48 || *(var + i) > 57) && *(var + i) != 32 && *(var + i) != 9 && *(var + i) != 45) return 0;
	}
return 1;
}
/* If NULL, value is not a number */
return 0;
}

/*************************************************************************************************/

void Print_Error(int no, ...)
{
va_list args;
/* Read arguments incoming in functionheader */
va_start(args, no);
/* Linux messages go to the console */
#ifdef LINUX
switch (no){
	case 1: fprintf(stderr, "%s", va_arg(args, char*)); break;
	case 2: fprintf(stderr, "%s", va_arg(args, char*)); printf("%d", va_arg(args, int)); break;
}
printf("\n");
#elif WIN32
char *data;
char str[10];

data = (char*)malloc(sizeof(char)*MAX_LEN);
strcpy(data, "");
switch (no) {
	case 1: strcat(data, va_arg(args, char*));
			break;
	case 2: strcat(data, va_arg(args, char*));
			sprintf(str, "%d", va_arg(args, int));
			strcat(data, str);
			break;
}
/* Windows messages with messagebox */
MessageBox(NULL, data, "GTK-server error", MB_OK | MB_ICONSTOP);
free(data);
#endif
va_end(args);
exit(-1);
}

/*************************************************************************************************/

#ifdef LINUX
void sig_handler(int signal)
{
int status;
/* A child process was killed, decrease fork counter */
(*count_fork)--;
/* Wait for child without blocking */
if (waitpid(-1, &status, WNOHANG) < 0) return;
}
#endif

/*************************************************************************************************/

void gtk_server_callback(GtkWidget *widget, gpointer *data)
{
Current_Object.object = widget;
Current_Object.state = (int)data;
Current_Object.text = NULL;
/* Future expansion */
Current_Object.mousex = 0;
Current_Object.mousey = 0;
}

/*************************************************************************************************/

void gtk_server_callback_extra(GtkWidget *widget, gpointer data)
{
Current_Object.object = widget;
Current_Object.state = (int)data;
Current_Object.text = (char*)data;
/* Future expansion */
Current_Object.mousex = 0;
Current_Object.mousey = 0;
}

/*************************************************************************************************/

#ifdef LINUX
void Void_Gtk(void *handle, void *extra_handle, char *name, char *amount, void *arg[], gfloat farg[], int model, char *buffer)
{
void (*func)();		/* In case of non-returning value */

/* Search function in GTK library */
*(void**)&func = dlsym(handle, name);
if (func == NULL) {
	*(void**)&func = dlsym(extra_handle, name);		/* Fix by Sebastiaan van Erk */
	if (func == NULL) {
		fprintf (stderr, "ERROR: GTK call not found in libraries \n%s\n", dlerror());
		exit(-1);
	}
}
#endif

#ifdef WIN32
void Void_Gtk(HINSTANCE hinstLib, HINSTANCE hLib, char *name, char *amount, void *arg[], gfloat farg[], int model, char *buffer)
{
FARPROC *func;
func = (FARPROC*)malloc(sizeof(FARPROC));

/* Search it function GTK library */
*func = (FARPROC)GetProcAddress(hinstLib, name);
if (*func == NULL) {
	*func = (FARPROC)GetProcAddress(hLib, name);
	if (*func == NULL){
		Print_Error (1, "ERROR: GTK call not found in library!");
	}
}
#endif

/* Realize function */
Void_Func(func, atoi(amount), model, arg, farg);

/* Free the malloced space after the call */
#ifdef WIN32
free(func);
#endif

/* Return positive response */
strcpy(buffer, "ok");
return;
}

/*************************************************************************************************/

#ifdef LINUX
void Widget_Gtk(void *handle, void *extra_handle, char *name, char *amount, char *type, int counter, void *arg[], gfloat farg[], int model, char *buffer)
{
GtkWidget *(*func)();	/* In case function returns a widget */
GtkWidget *widget;		/* Temporary widget holder */

/* Search function in GTK library */
*(void**)&func = dlsym(handle, name);
if (func == NULL) {
	*(void**)&func = dlsym(extra_handle, name);		/* Fix by Sebastiaan van Erk */
	if (func == NULL) {
		fprintf (stderr, "ERROR: GTK call not found in librarys \n%s\n", dlerror());
		exit(-1);
	}
}
#endif

#ifdef WIN32
void Widget_Gtk(HINSTANCE hinstLib, HINSTANCE hLib, char *name, char *amount, char *type, int counter, void *arg[], gfloat farg[], int model, char *buffer)
{
FARPROC *func;
GtkWidget *widget;		/* Temporary widget holder */

func = (FARPROC*)malloc(sizeof(FARPROC));

/* Search function in GTK library */
*func = (FARPROC)GetProcAddress(hinstLib, name);
if (*func == NULL) {
	*func = (FARPROC)GetProcAddress(hLib, name);
	if (*func == NULL){
		Print_Error (1, "ERROR: GTK call not found in library!");
	}
}
#endif

/* Realize function */
widget = Widget_Func(func, atoi(amount), model, arg, farg);

/* Free the malloced space after the call */
#ifdef WIN32
free(func);
#endif
/* Connect signals */
if (strcmp(type, "NONE")) gtk_signal_connect(GTK_OBJECT(widget), type, GTK_SIGNAL_FUNC(gtk_server_callback), (gpointer*)counter);
/* Convert to long value */
sprintf(buffer, "%d", (int)(widget));
/* Return positive response */
return;
}

/*************************************************************************************************/

#ifdef LINUX
void String_Gtk(void *handle, void *extra_handle, char *name, char *amount, void *arg[], gfloat farg[], int model, char *buffer)
{
char *(*func)();		/* In case function returns a string */
char *tmp;				/* temp result holder */

/* Search function in GTK library */
*(void**)&func = dlsym(handle, name);
if (func == NULL) {
	*(void**)&func = dlsym(extra_handle, name);		/* Fix by Sebastiaan van Erk */
	if (func == NULL) {
		fprintf (stderr, "ERROR: GTK call not found in librarys \n%s\n", dlerror());
		exit(-1);
	}
}
#endif

#ifdef WIN32
void String_Gtk(HINSTANCE hinstLib, HINSTANCE hLib, char *name, char *amount, void *arg[], gfloat farg[], int model, char *buffer)
{
FARPROC *func;
func = (FARPROC*)malloc(sizeof(FARPROC));
char *tmp;				/* temp result holder */

/* Search function in GTK library */
*func = (FARPROC)GetProcAddress(hinstLib, name);
if (*func == NULL) {
	*func = (FARPROC)GetProcAddress(hLib, name);
	if (*func == NULL){
		Print_Error (1, "ERROR: GTK call not found in library!");
	}
}
#endif

/* Realize function */
tmp = String_Func(func, atoi(amount), model, arg, farg);

/* Free the malloced space after the call */
#ifdef WIN32
free(func);
#endif
/* Now define result */
if(tmp != NULL){
	strcpy(buffer, tmp);
}
else strcpy(buffer, "");
/* Return */
return;
}
/*************************************************************************************************/

#ifdef LINUX
void Bool_Gtk(void *handle, void *extra_handle, char *name, char *amount, void *arg[], gfloat farg[], int model, char *buffer)
{
gboolean (*func)();		/* In case function returns a bool */
gboolean boolean;		/* Temporary bool holder */

/* Search function in GTK library */
*(void**)&func = dlsym(handle, name);
if (func == NULL) {
	*(void**)&func = dlsym(extra_handle, name);		/* Fix by Sebastiaan van Erk */
	if (func == NULL) {
		fprintf (stderr, "ERROR: GTK call not found in librarys \n%s\n", dlerror());
		exit(-1);
	}
}
#endif

#ifdef WIN32
void Bool_Gtk(HINSTANCE hinstLib, HINSTANCE hLib, char *name, char *amount, void *arg[], gfloat farg[], int model, char *buffer)
{
FARPROC *func;
gboolean boolean;		/* Temporary bool holder */

func = (FARPROC*)malloc(sizeof(FARPROC));

/* Search function in GTK library */
*func = (FARPROC)GetProcAddress(hinstLib, name);
if (*func == NULL) {
	*func = (FARPROC)GetProcAddress(hLib, name);
	if (*func == NULL){
		Print_Error (1, "ERROR: GTK call not found in library!");
	}
}
#endif

/* Realize function */
boolean = Bool_Func(func, atoi(amount), model, arg, farg);

/* Free the malloced space after the call */
#ifdef WIN32
free(func);
#endif
/* Convert to bool value */
if (boolean) strcpy(buffer, "1");
else strcpy(buffer, "0");
/* Return result */
return;
}

/*************************************************************************************************/

#ifdef LINUX
void Long_Gtk(void *handle, void *extra_handle, char *name, char *amount, void *arg[], gfloat farg[], int model, char *buffer)
{
guint (*func)();		/* In case function returns a long */
guint result;			/* Temporary result holder */

/* Search function in GTK library */
*(void**)&func = dlsym(handle, name);
if (func == NULL) {
	*(void**)&func = dlsym(extra_handle, name);		/* Fix by Sebastiaan van Erk */
	if (func == NULL) {
		fprintf (stderr, "ERROR: GTK call not found in librarys \n%s\n", dlerror());
		exit(-1);
	}
}
#endif

#ifdef WIN32
void Long_Gtk(HINSTANCE hinstLib, HINSTANCE hLib, char *name, char *amount, void *arg[], gfloat farg[], int model, char *buffer)
{
FARPROC *func;
guint result;			/* Temporary result holder */

func = (FARPROC*)malloc(sizeof(FARPROC));

/* Search function in GTK library */
*func = (FARPROC)GetProcAddress(hinstLib, name);
if (*func == NULL) {
	*func = (FARPROC)GetProcAddress(hLib, name);
	if (*func == NULL){
		Print_Error (1, "ERROR: GTK call not found in library!");
	}
}
#endif

/* Realize function */
result = Long_Func(func, atoi(amount), model, arg, farg);

/* Free the malloced space after the call */
#ifdef WIN32
free(func);
#endif
/* Convert to long value */
sprintf(buffer, "%d", result);
/*  Return positive response */
return;
}

/*************************************************************************************************/

#ifdef LINUX
void Float_Gtk(void *handle, void *extra_handle, char *name, char *amount, void *arg[], gfloat farg[], int model, char *buffer)
{
gfloat (*func)();		/* In case function returns a float */
gfloat result;			/* Temporary result holder */

/* Search function in GTK library */
*(void**)&func = dlsym(handle, name);
if (func == NULL) {
	*(void**)&func = dlsym(extra_handle, name);		/* Fix by Sebastiaan van Erk */
	if (func == NULL) {
		fprintf (stderr, "ERROR: GTK call not found in librarys \n%s\n", dlerror());
		exit(-1);
	}
}
#endif

#ifdef WIN32
void Float_Gtk(HINSTANCE hinstLib, HINSTANCE hLib, char *name, char *amount, void *arg[], gfloat farg[], int model, char *buffer)
{
FARPROC *func;
gfloat result;			/* Temporary result holder */

func = (FARPROC*)malloc(sizeof(FARPROC));

/* Search function in GTK library */
*func = (FARPROC)GetProcAddress(hinstLib, name);
if (*func == NULL) {
	*func = (FARPROC)GetProcAddress(hLib, name);
	if (*func == NULL){
		Print_Error (1, "ERROR: GTK call not found in library!");
	}
}
#endif

/* Realize function */
result = Float_Func(func, atoi(amount), model, arg, farg);

/* Free the malloced space after the call */
#ifdef WIN32
free(func);
#endif
/* Convert to float value */
sprintf(buffer, "%f", result);
/*  Return */
return;
}

/*************************************************************************************************/

/* This is the argument parser created by Jim Bailey, adjusted slightly to fit into the GTK-server. */

GList *parse_line(char *data)
{
	GList *args = NULL;
	int start_idx = 0;
	int curr_idx = 0;
	int quote_dst_idx = 0;
	int state = 0;
	char c;

	const int in_plain     = 1 << 1;
	const int in_quotes    = 1 << 2;
	const int after_escape = 1 << 3;

	/* loop over the contents */
	for ( ; curr_idx < strlen(data) + 1; ++curr_idx )
	{
		c = data[curr_idx];

		if ( state & in_plain )
		{
			if ( c == ' ' || c == '\t' || c == '\n' || c == '\r' || curr_idx == strlen(data))
			{
				args = g_list_append(args, g_strndup(data + start_idx, curr_idx - start_idx ) );
				state ^= in_plain;
			}
		}
		else if ( state & in_quotes )
		{
			if ( state & after_escape )
			{
				if ( c == 'n' )
					c = '\n';
				else if ( c == 't' )
					c = '\t';
				else if ( c == 'r' )
					c = '\r';
				
				data[quote_dst_idx] = c;
				++quote_dst_idx;

				state ^= after_escape;
			}
			else if ( c == '"' )
			{
				args = g_list_append(args, g_strndup(data + start_idx, quote_dst_idx - start_idx ) );
				state ^= in_quotes;
			}
			else if ( c == '\\' )
			{
				state |= after_escape;
			}
			else
			{
				data[quote_dst_idx] = c;
				++quote_dst_idx;
			}
		}
		else
		{
			if ( c == '"' )
			{
				state |= in_quotes;
				start_idx = quote_dst_idx = curr_idx + 1;
			}
			else if ( c != ' ' && c != '\t' && c != '\n' && c != '\r' )
			{
				state |= in_plain;
				start_idx = curr_idx;
			}
		}
	}
	return (args);
}

/*************************************************************************************************/

#ifdef LINUX
void Call_Realize (char *inputdata, CONFIG *Find_Config, void *handle, void* extra_handle, char *retstr)
#endif
#ifdef WIN32
void Call_Realize (char *inputdata, CONFIG *Find_Config, HINSTANCE handle, HINSTANCE extra_handle, char *retstr)
#endif
{
char *gtk_api_call;
char *arg;				/* Temporary argument holders */
GList *list;			/* This will contain the list with individual arguments */
char *widget;
char *signal;
void *tmparg[10];		/* Array needed for arguments of the API call */
int i;
gfloat farg[10];		/* Array to keep the float arguments */
int model;				/* Keeps track of function model */

if (strlen(inputdata) > 0) {
	/* Get the API call part */
	list = parse_line(inputdata);
	gtk_api_call = g_list_nth_data(list, 0);
	/* Have we found the GTK-signal call? */
	if (!strcmp("gtk_server_callback", gtk_api_call)) {
		/* Yes, find the first argument */
		if ((arg = g_list_nth_data(list, 1)) != NULL){
			while (1){
				/* Let GTK wait for an event */
				if (!strcmp (arg, "WAIT") || !strcmp (arg, "wait") || !strcmp(arg, "1")) {gtk_main_iteration();}
				/* Update asynchronously */
				if (!strcmp (arg, "UPDATE") || !strcmp (arg, "update") || !strcmp(arg, "2")) {while (gtk_events_pending()) gtk_main_iteration();}
				/* Find out the callback state for the given object for all arguments to gtk_server_callback */
				if ((long)Current_Object.state != GTK_SERVER_NONE) {
					/* Signal found */
					Current_Object.state = GTK_SERVER_NONE;
					/* Convert long value containing widget ref. to string */
					if (Current_Object.text == NULL) sprintf(retstr, "%d", (int)Current_Object.object);
					/* Or return self-defined string */
					else strcpy(retstr, Current_Object.text);
					/* Free memory of members and list */
					g_list_foreach(list, (GFunc)g_free, NULL);
					g_list_free(list);
					return;
				}
				/* Only return with '0' to client if argument was not a WAIT argument */
				if (strcmp (arg, "WAIT") && strcmp (arg, "wait") && strcmp(arg, "1")){
					strcpy (retstr, "0");
					/* Free memory of members and list */
					g_list_foreach(list, (GFunc)g_free, NULL);
					g_list_free(list);
					return;
				}
			}
		}
		/* GTK-SERVER call found, but no argument given */
		else {
			Print_Error(1, "ERROR: Missing WAIT state GTK_SERVER_CALLBACK!");
		}
	}
	/* Internal call for version */
	if (!strcmp("gtk_server_version", gtk_api_call)){
		/* Return GTK-server version */
		strcpy(retstr, "GTK-server ");
		strcat(retstr, VERSION);
		/* Free memory of members and list */
		g_list_foreach(list, (GFunc)g_free, NULL);
		g_list_free(list);
		return;
	}
	/* Internal call for returning string with particular callback */
	if (!strcmp("gtk_server_connect", gtk_api_call)){
		if ((widget = g_list_nth_data(list, 1)) == NULL){
			Print_Error(1, "ERROR: Cannot find widget reference in GTK_SERVER_CONNECT!");
		}
		if ((signal = g_list_nth_data(list, 2)) == NULL){
			Print_Error(1, "ERROR: Cannot find signal type in GTK_SERVER_CONNECT!");
		}
		if ((arg = g_list_nth_data(list, 3)) == NULL){
			Print_Error(1, "ERROR: Cannot find response string in GTK_SERVER_CONNECT!");
		}
		gtk_signal_connect(GTK_OBJECT((GtkObject*)(atol(widget))), Trim_String(signal), GTK_SIGNAL_FUNC(gtk_server_callback_extra), (gpointer)(strdup(arg)));
		strcpy(retstr, "ok");
		/* Free memory of members and list */
		g_list_foreach(list, (GFunc)g_free, NULL);
		g_list_free(list);
		return;
	}
	/* No GTK-SERVER call, so check for a genuine GTK call
		Check GTK_CONFIG list */
	while ((Find_Config != NULL) && (strcmp(Find_Config->name, gtk_api_call))) {
		Find_Config = Find_Config->next;
	}
	/* Have we found a GTK call? */
	if (Find_Config != NULL) {
		/* Put model of function to 0 */
		model = 0;
		/* Yes, now get the arguments if there are any */
		if(atoi(Find_Config->argamount) > 0){
			/* Make sure to read '.' as RADIX */
			#ifdef LINUX
			setlocale(LC_NUMERIC, "POSIX");
			#elif WIN32
			setlocale(LC_NUMERIC, "English");
			#endif
			/* Determine type of argument */
			for (i = 0; i < atoi(Find_Config->argamount); i++) {
				arg = g_list_nth_data(list, i+1);
				model = model << 1;
				if (!strcmp(Find_Config->args[i], "LONG")){
					/* Check if the argument really is a number? */
					if (is_value (arg)) tmparg[i] = (long*)atol(arg);
					/* If not a number, then it is a string */
					else tmparg[i] = (char*)arg;
				}
				if (!strcmp(Find_Config->args[i], "WIDGET")) {
					/* Check if the value is a number > 0 */
					if (atol(arg) > 0 || !strcmp(arg, "NULL")) tmparg[i] = (GtkObject*)(atol(arg));
					else Print_Error(1, "Cannot cast argument to widget!");
				}
				if (!strcmp(Find_Config->args[i], "STRING")) tmparg[i] = (char*)arg;
				if (!strcmp(Find_Config->args[i], "FLOAT")){
					farg[i] = (float)atof(arg);
					model += 1;
				}
				if (!strcmp(Find_Config->args[i], "NULL")) tmparg[i] = NULL;
			}
		}
		/* The GTK CALL has no return value */
		if (!strcmp(Find_Config->returnvalue, "NONE"))
			Void_Gtk(handle, extra_handle, Find_Config->name, Find_Config->argamount, tmparg, farg, model, retstr);
		/* The GTK CALL has a widget return value */
		else if (!strcmp(Find_Config->returnvalue, "WIDGET"))
			Widget_Gtk(handle, extra_handle, Find_Config->name, Find_Config->argamount, Find_Config->callbacktype, Find_Config->counter, tmparg, farg, model, retstr);
		/* The GTK CALL delivers a string as a returnvalue */
		else if (!strcmp(Find_Config->returnvalue, "STRING"))
			String_Gtk(handle, extra_handle, Find_Config->name, Find_Config->argamount, tmparg, farg, model, retstr);
		/* The GTK CALL delivers a boolean as a returnvalue */
		else if (!strcmp(Find_Config->returnvalue, "BOOL"))
			Bool_Gtk(handle, extra_handle, Find_Config->name, Find_Config->argamount, tmparg, farg, model, retstr);
		/* The GTK CALL delivers an INTEGER as a returnvalue */
		else if (!strcmp(Find_Config->returnvalue, "LONG"))
			Long_Gtk(handle, extra_handle, Find_Config->name, Find_Config->argamount, tmparg, farg, model, retstr);
		/* The GTK CALL delivers a FLOAT as a returnvalue */
		else if (!strcmp(Find_Config->returnvalue, "FLOAT"))
			Float_Gtk(handle, extra_handle, Find_Config->name, Find_Config->argamount, tmparg, farg, model, retstr);
		/* Other return value (future expansion) */
		else{
		}
		/* Free memory of members and list */
		g_list_foreach(list, (GFunc)g_free, NULL);
		g_list_free(list);
	}
	/* API call not found */
	else strcpy(retstr, "-1");
}
return;
}

/*************************************************************************************************
************************************************************* This is the main loop of the server
*************************************************************************************************/

int main(int argc, char *argv[])
{
/* Define list for CONFIGURATION settings */
CONFIG *Gtk_Api_Config;
CONFIG *Conf_Last;
CONFIG *Start_Gtk_Api_Config;
/* Define the configfile */
FILE *configfile;
/* Define the line input, declare memory */
char *line = (char*)malloc(MAX_LEN*sizeof(char));
/* Define the libraryname to open */
char *libname; 
char *extra_libname; 
/* Define environment variable holder */ 
char *libvar;
/* Count current line number of configfile */
int count_line;
/* Define handle to GTK library */
#ifdef LINUX
char *filename;
void *handle;
void *extra_handle;
socklen_t sin_size, addr_len;
#endif
#ifdef WIN32
HINSTANCE handle;
HINSTANCE extra_handle;
/* Needed for OS detection */
OSVERSIONINFOEX osvi;
BOOL bOsVersionInfoEx;
/* Needed for PIPES */
HANDLE hPipe1;
HANDLE hPipe2;
CHAR chRequest[MAX_LEN]; 
DWORD cbBytesRead, cbWritten; 
BOOL fSuccess;
char *dialog_msg = (char*)malloc(MAX_LEN*sizeof(char));
/* Needed for finding the configfile */
char filename[MAX_PATH];
int len;
int sin_size, addr_len;
#endif
/* Define temp variable for loops */
int i;
/* This pointer contains the string which is returned to the client program */
char *retstr;
/* Define vars for TCP sockets */
char *host;
char *port;
char *maxtcp;
int sockfd;
struct hostent *he;
struct sockaddr_in my_addr;                 /* my address information */
struct sockaddr_in their_addr;				/* connector's address information (UDP) */
struct hostent *se;							/* Needed for UDP */
int new_fd;
int yes = 1;
int numbytes;
char buf[MAX_LEN];							/* Buffer containing data from socket or file */
/* Define logfile facility */
FILE *log;
char *LogDir;
#ifdef LINUX
/* SIGNAL struct for child processes */
struct sigaction sa;
int curpid;									/* Needed for forking */
count_fork = (int*)malloc(sizeof(int));
#endif

/* Run help if there are no arguments */
if (argc < 2) {
	#ifdef WIN32
	strcpy(dialog_msg, "*** This is the GTK-server ");
	strcat(dialog_msg, VERSION);
	strcat(dialog_msg, " *** \
		\n\r\n\rUsage:\tgtk-server = this help \
		\n\r\tgtk-server stdin [log] = start server in STDIN mode \
		\n\r\tgtk-server fifo [log] = start server in FIFO mode \
		\n\r\tgtk-server tcp <host:port> [log] = start server in TCP mode \
		\n\r\tgtk-server udp <host:port> [log] = start server in UDP mode");
	MessageBox(NULL, dialog_msg, "GTK-server usage", MB_OK | MB_ICONINFORMATION);
	exit(0);
	#endif
	#ifdef LINUX
	printf("\n*** This is the GTK-SERVER ");
	printf("%s ", VERSION);
	#ifdef GTK1x
	printf("compiled for GTK 1.x. ***\n\n");
	#elif GTK2x
	printf("compiled for GTK 2.x. ***\n\n");
	#endif
	printf("Usage:");
	printf("\tgtk-server = this help\n");
	printf("\tgtk-server stdin [log] = start server in STDIN mode\n");
	printf("\tgtk-server fifo <name> [log] = start server in FIFO mode\n");
	printf("\tgtk-server tcp <host:port[:max]> [log] = start server in TCP mode\n");
	printf("\tgtk-server udp <host:port> [log] = start server in UDP mode\n\n");
	exit (0);
	#endif
}

/* Open the config file, first in the same directory as the script */
configfile = fopen ("gtk-server.cfg", "r");
if (configfile == NULL){
	/* Try to find the environment variable GTK_SERVER_CONFIG */

	libvar = getenv("GTK_SERVER_CONFIG");
	configfile = fopen (libvar, "r");
	#ifdef WIN32
	if (configfile == NULL) {
		/* Now look for configfile in same directory as gtk-server.exe ('me') */
		GetModuleFileName(NULL, filename, MAX_PATH);
		PathRemoveExtension(filename);
		len = strlen(filename);
		filename[len] = '.';
		filename[++len] = 'c';
		filename[++len] = 'f';
		filename[++len] = 'g';
		filename[++len] = '\0';
		configfile = fopen(filename, "r");
	}
	#endif
	#ifdef LINUX
	/* Find a configfile in the default Linux directory's */
	if (configfile == NULL) {
		/* Homedir */
		libvar = getenv("HOME"); 
		if (libvar != NULL){
			filename = (char *)malloc(strlen(libvar) + 17);		/* Fix by Sebastiaan van Erk */
			strcpy(filename, libvar);
			strcat(filename, "/.gtk-server.cfg");
			configfile = fopen (filename, "r");
			free(filename);
        }
		if (configfile == NULL || libvar == NULL){
			configfile = fopen ("/etc/gtk-server.cfg", "r");
			if (configfile == NULL){
				configfile = fopen ("/usr/local/etc/gtk-server.cfg", "r");
			}
		}
	}
	#endif
}

/* No configfile found? Exit */
if (configfile == NULL) Print_Error(1, "ERROR: No 'gtk-server.cfg' file found!");

count_line = 0;

/* Claim memory for configfile entrys */
Gtk_Api_Config = (CONFIG*)malloc(sizeof(CONFIG));
Start_Gtk_Api_Config = Gtk_Api_Config;
/* Claim memory for LOGDIR definition */
LogDir = (char*)malloc(MAX_LEN*sizeof(char));

/* Extract the entry's of the configfile into the array */
while (fgets (line, MAX_LEN, configfile) != NULL){
	count_line++;
	/* Check if the line is a comment or the line starts with spaces */
	if (strncmp(line, "#", 1) && strlen(line) > 1) {
		/* No, check if this is the gtklibname config setting */
		if (!strncmp(line, "GTK_LIB_NAME", 12)) {
			strtok(line, "=");
			libname = Trim_String(strtok(NULL, "="));
			if (strlen(libname) <= 1) Print_Error(2, "ERROR: Missing gtk library name in GTK config file at line: ", count_line);
		}
		/* Is there an extra library? */
		else if (!strncmp(line, "GTK_LIB_EXTRA", 13)){
			strtok(line, "=");
			extra_libname = Trim_String(strtok(NULL, "="));
		}
		/* Check if we have a GTK function definition */
		else if (!strncmp(line, "GTK_LIB_FUNCTION", 16)){
			strtok(line, "=");
			Gtk_Api_Config->name = strtok(NULL, ",");
			if (Gtk_Api_Config->name != NULL) {
				/* Get rid of surrounding whitespaces */
				Gtk_Api_Config->name = Trim_String(Gtk_Api_Config->name);
				/* Get next term: callback signal type */
				if ((Gtk_Api_Config->callbacktype = strtok(NULL, ",")) != NULL){
					Gtk_Api_Config->callbacktype = Trim_String(Gtk_Api_Config->callbacktype);
					/* Determine a unique number used for internal identification - I use the current line in the configfile */
					Gtk_Api_Config->counter = count_line;
				}
				else Print_Error(2, "ERROR: Missing callbacktype in GTK config file at line: ", count_line);
				/* Get next term: return value */
				if ((Gtk_Api_Config->returnvalue = strtok(NULL, ",")) != NULL) Gtk_Api_Config->returnvalue = Trim_String(Gtk_Api_Config->returnvalue);
				else Print_Error(2, "ERROR: Missing returnvalue in GTK config file at line: ", count_line);
				/* Get next term: amount of arguments */
				if ((Gtk_Api_Config->argamount = strtok(NULL, ",")) != NULL) Gtk_Api_Config->argamount = Trim_String(Gtk_Api_Config->argamount);
				else Print_Error(2, "ERROR: Missing argumentamount in GTK config file at line: ", count_line);
				/* Get the separate arguments */
				for (i = 0; i < atoi(Gtk_Api_Config->argamount); i++){
					if((Gtk_Api_Config->args[i] = strtok(NULL, ",")) != NULL) Gtk_Api_Config->args[i] = Trim_String(Gtk_Api_Config->args[i]);
				}
			}
			/* Declare next item */
			Conf_Last = Gtk_Api_Config;
			Gtk_Api_Config->next = (CONFIG*)malloc(sizeof(CONFIG));
			Gtk_Api_Config = Gtk_Api_Config->next;
			Gtk_Api_Config->next = NULL;
		}
		/* Check if we have a LOGDIR definition */
		else if (!strncmp(line, "GTK_SERVER_LOG", 14)){
			strtok(line, "=");
			LogDir = Trim_String(strtok(NULL, "="));
		}
		else Print_Error(2, "ERROR: Config not recognized at line: ", count_line);
		/* Claim memory for next config entry */
		line = (char*)malloc(MAX_LEN*sizeof(char));
	}
}
/* No records left, define end points */
Conf_Last->next = NULL;

/* Make sure there is a logging dir */
if (!strcmp(LogDir, "")) strcpy(LogDir, ".");

/* If last argument is 'log' then create logfile for writing */
strcat(LogDir, "/gtk-server.log");
/* For Linux, only create logfile when argument is given */
#ifdef LINUX
if (!strcmp(argv[argc - 1], "log")){
	log = fopen (LogDir, "w");
	if (log == NULL) printf("WARNING: The logfile could not be created. Please check your configfile.\n");
}
#endif
/* For Windows, always create a logfile to capture GTK warnings */
#ifdef WIN32
log = freopen (LogDir, "w", stderr);
if (log == NULL) MessageBox(NULL, "WARNING: The logfile could not be created. Please check your configfile.\n\rGTK errors will appear in a DOS box.", "GTK-server warning", MB_OK | MB_ICONEXCLAMATION);
#endif

/* Open the GTK library */
#ifdef SHARED
handle = dlopen(libname, RTLD_LAZY);
if (!handle){
	printf("ERROR: The GTK library could not be opened.\n");
	fputs(dlerror(), stderr);
	exit(-1);
}
if (extra_libname != NULL) extra_handle = dlopen(extra_libname, RTLD_LAZY);
#endif

#ifdef STATIC
handle = dlopen(NULL, RTLD_LAZY);
if (!handle){
	printf("ERROR: The GTK library could not be opened.\n");
	fputs(dlerror(), stderr);
	exit(-1);
}
if (extra_libname != NULL) extra_handle = dlopen(extra_libname, RTLD_LAZY);
#endif

#ifdef WIN32
handle = LoadLibrary(libname);
if (handle == NULL) Print_Error(2, "ERROR: The GTK library could not be opened - ", (int)GetLastError());
if (extra_libname != NULL) extra_handle = LoadLibrary(extra_libname);
#endif

/* Claim memory for returnvalue */
retstr = (char*)malloc(MAX_LEN*sizeof(char));

/**************************************************************************************** STDIN */
/* There is an argument, check on STDIN flag */
if (!strcmp(argv[1], "stdin")) {
		
	/* This is the main input loop via STDIN */
	while (1) {
		/* Perform a low-level READ and expect more than 1 byte */
		do {
			i = read(STDIN, line, MAX_LEN);
			if (i == -1){
				perror("read");		/* Fix by Sebastiaan van Erk */
				exit(-1);
			}
			if (i == 0) exit(0);
		} while (i < 2); 

		line[i] = '\0';

		/* If logging is enabled, write incoming text to log */
		if (argc == 3 && log != NULL){
			fprintf(log, "SCRIPT: %s\n", line);
			fflush(log);
		}
		Call_Realize(Trim_String(line), Start_Gtk_Api_Config, handle, extra_handle, retstr);
		/* If logging is enabled, write returnstring to log */
		if (argc == 3 && log != NULL){
			fprintf(log, "SERVER: %s\n", retstr);
			fflush(log);
		}
		printf("%s\n", retstr);
		fflush(stdout);
	}
}

#ifdef LINUX
/**************************************************************************************** Linux FIFO */
/* Check on the FIFO flag */
if (!strcmp(argv[1], "fifo")) {
	/* Create FIFO file */
	if (mkfifo(argv[2], S_IWUSR | S_IRUSR | S_IWGRP | S_IRGRP | S_IWOTH | S_IROTH) < 0) {
		perror("mkfifo");
		exit(-1);
	}
	/* This is the main input loop via FIFO */
	while (1) {
		/* First open in READ mode */
		if((sockfd = open(argv[2], O_RDONLY)) < 0){
			perror("open"); exit(-1);
		}
		/* Now wait for data */
		if((numbytes = read(sockfd, buf, MAX_LEN-1)) < 0){
			perror("read"); exit(-1);
		}
		buf[numbytes] = '\0';
		/* Close socket again */
		if(close(sockfd) < 0){
			perror("close"); exit(-1);
		}
		/* If logging is enabled, write incoming text to log */
		if (argc == 4 && log != NULL){
			fprintf(log, "SCRIPT: %s\n", Trim_String(buf));
			fflush(log);
		}
		Call_Realize(Trim_String(buf), Start_Gtk_Api_Config, handle, extra_handle, retstr);

		/* If logging is enabled, write returnstring to log */
		if (argc == 4 && log != NULL){
			fprintf(log, "SERVER: %s\n", retstr);
			fflush(log);
		}

		/* Add newline to returned string */
		strcat(retstr, "\n");

		/* Now open in WRITE mode */
		if((sockfd = open(argv[2], O_WRONLY | O_SYNC)) < 0){
			perror("open"); exit(-1);
		}
		/* Write data */
		if(write (sockfd, retstr, strlen(retstr)) < 0){
			perror("write"); exit(-1 );
		}
		/* Close socket again */
		if(close(sockfd) < 0){
			perror("close"); exit(-1);
		}
	}
}
#endif

#ifdef WIN32
/**************************************************************************************** Win32 FIFO */
/* Check for a named PIPE */
if (!strcmp(argv[1], "fifo")) {
	
	/* Checkh the Operating System version first */
	ZeroMemory(&osvi, sizeof(OSVERSIONINFOEX));
	osvi.dwOSVersionInfoSize = sizeof(OSVERSIONINFOEX);

	if( !(bOsVersionInfoEx = GetVersionEx ((OSVERSIONINFO *) &osvi))) {
		osvi.dwOSVersionInfoSize = sizeof (OSVERSIONINFO);
		if (! GetVersionEx ((OSVERSIONINFO *) &osvi)) return FALSE;
	}
	
	/* If windows9x/ME then block PIPE function */
	if (osvi.dwPlatformId == VER_PLATFORM_WIN32_WINDOWS) Print_Error(1, "ERROR: Named pipes are available on WINNT, WIN2000 and WINXP only!");
	
	/* Now create named PIPE - code ripped from msdn.microsoft.com */
	hPipe1 = CreateNamedPipe( 
		"\\\\.\\pipe\\out",         /* pipe name */
		PIPE_ACCESS_INBOUND,       /* read/write access */
		PIPE_TYPE_MESSAGE |       /* message type pipe */
		PIPE_READMODE_MESSAGE |   /* message-read mode */
		PIPE_WAIT,                /* blocking mode */
		PIPE_UNLIMITED_INSTANCES, /* max. instances */
		MAX_LEN,                  /* output buffer size */
		MAX_LEN,                  /* input buffer size */
		NMPWAIT_USE_DEFAULT_WAIT, /* client time-out */
		NULL);                    /* no security attribute */

	/* Exit if pipe cannot be created */
	if (hPipe1 == INVALID_HANDLE_VALUE) Print_Error(1, "Could not create a named pipe!"); 
	
	/* Now create named PIPE - code ripped from msdn.microsoft.com */
	hPipe2 = CreateNamedPipe( 
		"\\\\.\\pipe\\in",        /* pipe name */
		PIPE_ACCESS_OUTBOUND,     /* read/write access */
		PIPE_TYPE_MESSAGE |       /* message type pipe */
		PIPE_READMODE_MESSAGE |   /* message-read mode */
		PIPE_WAIT,                /* blocking mode */
		PIPE_UNLIMITED_INSTANCES, /* max. instances */
		MAX_LEN,                  /* output buffer size */
		MAX_LEN,                  /* input buffer size */
		NMPWAIT_USE_DEFAULT_WAIT, /* client time-out */
		NULL);                    /* no security attribute */

	/* Exit if pipe cannot be created */
	if (hPipe2 == INVALID_HANDLE_VALUE) Print_Error(1, "Could not create a named pipe!\n"); 
	
	/* Wait for connection from client script */
	if (ConnectNamedPipe(hPipe1, NULL)) {
		if (ConnectNamedPipe(hPipe2, NULL)) {

		/* This is the main input loop via FIFO */
		while(1) {

			/* Connection established, now communicate */
			fSuccess = ReadFile( 
			hPipe1,			/* handle to pipe */
			chRequest,		/* buffer to receive data */
			MAX_LEN,		/* size of buffer */
			&cbBytesRead,	/* number of bytes read */
			NULL);			/* not overlapped I/O */

			if (! fSuccess || cbBytesRead == 0) {
				Print_Error(2, "ERROR: GetLastError returned: ", (int)GetLastError());
				FlushFileBuffers(hPipe1);
				DisconnectNamedPipe(hPipe1);
				CloseHandle(hPipe1);
				exit(-1);
			}
			
			chRequest[cbBytesRead] = '\0';

			/* If logging is enabled, write incoming text to log */
			if (argc == 3 && log != NULL){
				fprintf(log, "SCRIPT: %s\n", chRequest);
				fflush(log);
			}

			Call_Realize(Trim_String(chRequest), Start_Gtk_Api_Config, handle, extra_handle, retstr);

			/* If logging is enabled, write returnstring to log */
			if (argc == 3 && log != NULL){
				fprintf(log, "SERVER: %s\n", retstr);
				fflush(log);
			}
			
			/* Add newline to returned string */
			strcat(retstr, "\n");
			
			/* Write the reply to the pipe. */
			fSuccess = WriteFile( 
			hPipe2,			/* handle to pipe */
			retstr,			/* buffer to write from */
			strlen(retstr),	/* number of bytes to write */
			&cbWritten,		/* number of bytes written */
			NULL);			/* not overlapped I/O */

			if (! fSuccess || strlen(retstr) != cbWritten) {
				Print_Error(2, "ERROR: GetLastError returned: ", (int)GetLastError());
				FlushFileBuffers(hPipe2);
				DisconnectNamedPipe(hPipe2);
				CloseHandle(hPipe2);
				exit(-1);
			}
		}
		}
	
	 	else {
			/* The client could not connect, so close the pipe. */
			CloseHandle(hPipe2);
			Print_Error(1, "Connection pipe 'in' refused!");
	 	}
	}
	else {
		/* The client could not connect, so close the pipe. */
		CloseHandle(hPipe1);
		Print_Error(1, "Connection pipe 'out' refused!");
	}
}
#endif

/**************************************************************************************** TCP */
/* Else consider this to be TCP socket - host:port */
if (!strcmp(argv[1], "tcp")) {

/*Check argument format */
if (strstr(argv[2], ":") == NULL) Print_Error(1, "ERROR: Argument is in wrong format!");

host = Trim_String(strtok(argv[2], ":"));
port = Trim_String(strtok(NULL, ":"));
#ifdef LINUX
maxtcp = strtok(NULL, ":");
#endif
if (maxtcp == NULL){
	maxtcp = (char*)malloc(8*sizeof(char));
	strcpy(maxtcp, "1");
}

sockfd = 0;
#ifdef WIN32
/* Initialize Win32 environment for sockets */
WSADATA wsaData;
if (WSAStartup(MAKEWORD(2, 0), &wsaData) != 0) Print_Error(1, "WSAStartup() failed!");
#endif
/* get the host info */
if ((he=gethostbyname(host)) == NULL){
	Print_Error (2, "Specified host does not exist!\n\nSystem message: ", errno);
}
else if ((sockfd = socket(AF_INET, SOCK_STREAM, 0)) == -1){
	Print_Error(2, "Cannot connect to socket!\n\nSystem message: ", errno);
}
#ifdef LINUX
else if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, &yes, sizeof(int)) == -1)
#endif
#ifdef WIN32
else if (setsockopt(sockfd, SOL_SOCKET, SO_REUSEADDR, (char*)&yes, sizeof(int)) == -1)
#endif
{
	Print_Error(2, "Cannot configure socket!\n\nSystem message: ", errno);
}
/* Perform BIND to the host:port */
else {
	my_addr.sin_family = AF_INET;
	my_addr.sin_port = htons((long)atol(port));             /* Fill in specified port */
	my_addr.sin_addr = *((struct in_addr *)he->h_addr);     /* Fill in specified IP address */
	memset(&(my_addr.sin_zero), '\0', 8);                   /* zero the rest of the struct */

	if (bind(sockfd, (struct sockaddr *)&my_addr, sizeof(struct sockaddr)) == -1) {
		Print_Error(2, "Unable to bind the specified socket address!\n\nSystem message: ", errno);
	}

	else if (listen(sockfd, BACKLOG) == -1) {
		Print_Error(2, "Unable to listen to the specified socket address!\n\nSystem message: ", errno);
	}
}
#ifdef LINUX
if (atoi(maxtcp) > 1) {
	/* Register zombie process reaper */
	sa.sa_handler = sig_handler;
	/* Do not block other signals */
	sigemptyset(&sa.sa_mask);
	sa.sa_flags = SA_RESTART;
	if (sigaction(SIGCHLD, &sa, NULL) < 0) {
		printf("Could not set signal handler!\n");
		perror("sigaction");
		exit(-1);
	}
}
/* No forks yet */
*count_fork = 0;
#endif
/* Now accept incoming connections */
while (1){
	sin_size = sizeof(struct sockaddr_in);
	if ((new_fd = accept(sockfd, (struct sockaddr *)&their_addr, &sin_size)) == -1){
		perror("accept");
		continue;
	}
	#ifdef WIN32
	closesocket(sockfd);
	#endif
	#ifdef LINUX
	/* Only one client may connect? No -> fork */
	if (atoi(maxtcp) > 1){
		/* Max not reached? Fork */
		if ((*count_fork) < atoi(maxtcp)){
			curpid = fork();
			/* Count connect */
			(*count_fork)++;
		}
	}
	else curpid = 0;
	/* We are in the non-forked process or in the child process */
	if (!curpid){
		/* Close the listener socket */
		close (sockfd);
		#endif
		/* We enter the mainloop - read incoming text */
		while(1){
			if ((numbytes = recv(new_fd, buf, MAX_LEN - 1, 0)) > 0){
				buf[numbytes] = '\0';

				/* If logging is enabled, write incoming text to log */
				if (argc == 4 && log != NULL){
					fprintf(log, "SCRIPT: %s\n", Trim_String((char*)buf));
					fflush(log);
				}
				Call_Realize(Trim_String((char*)buf), Start_Gtk_Api_Config, handle, extra_handle, retstr);

				/* If logging is enabled, write returnstring to log */
				if (argc == 4 && log != NULL){
					fprintf(log, "SERVER: %s\n", retstr);
					fflush(log);
				}
				/* Add newline to returned string */
				strcat(retstr, "\n");

				/* Now send the result back to the socket */
				send(new_fd, retstr, strlen(retstr), 0);
			}
			/* If disconnected, quit the gtk-server */
			else {
				exit(0);
			}
		}
	#ifdef LINUX
	}
	/* This is the parent process in case the server was forked */
	else {
		/* Close the new formed socket, the parent process does not need it */
		close (new_fd);
	}
	#endif
} /* Endless while */
} /* End TCP story */

/**************************************************************************************** UDP */
/* Else check if this to be UDP socket - host:port */
if (!strcmp(argv[1], "udp")) {

/* Check argument format */
if (strstr(argv[2], ":") == NULL) Print_Error(1, "ERROR: Argument is in wrong format!");
/* Find host IP address and port */
host = Trim_String(strtok(argv[2], ":"));
port = Trim_String(strtok(NULL, ":"));
sockfd = 0;
#ifdef WIN32
/* Initialize Win32 environment for sockets */
WSADATA wsaData;
if (WSAStartup(MAKEWORD(2, 0), &wsaData) != 0) Print_Error(1, "WSAStartup() failed!");
#endif

/* Get the host info */
if ((he=gethostbyname(host)) == NULL){
	Print_Error (2, "Specified host does not exist!\n\nSystem message: ", errno);
}
else if ((sockfd = socket(AF_INET, SOCK_DGRAM, 0)) == -1){
	Print_Error(2, "Cannot connect to socket!\n\nSystem message: ", errno);
}

/* Perform BIND to the host:port */
else {
	my_addr.sin_family = AF_INET;
	my_addr.sin_port = htons((long)atol(port));             /* Fill in specified port */
	my_addr.sin_addr = *((struct in_addr *)he->h_addr);     /* Fill in specified IP address */
	memset(&(my_addr.sin_zero), '\0', 8);                   /* zero the rest of the struct */

	if (bind(sockfd, (struct sockaddr *)&my_addr, sizeof(struct sockaddr)) == -1) {
		Print_Error(2, "Unable to bind the specified socket address!\n\nSystem message: ", errno);
	}
}

addr_len = sizeof(struct sockaddr);

/* Now accept incoming packets */
while (1){
	if ((numbytes=recvfrom(sockfd, buf, MAX_LEN - 1, 0, (struct sockaddr *)&their_addr, &addr_len)) > 0) {

		/* Terminate incoming string */
		buf[numbytes] = '\0';

		/* Only determine values when empty */
		if (se == NULL) {
			se = gethostbyname(inet_ntoa(their_addr.sin_addr));
			their_addr.sin_family = AF_INET;      /* host byte order */
			their_addr.sin_port = htons((long)atol(port));  /* short, network byte order */
			their_addr.sin_addr = *((struct in_addr *)se->h_addr);
			memset(&(their_addr.sin_zero), '\0', 8);                   /* zero the rest of the struct */
		}

		/* If logging is enabled, write incoming text to log */
		if (argc == 4 && log != NULL){
			fprintf(log, "SCRIPT: %s\n", Trim_String((char*)buf));
			fflush(log);
		}
		Call_Realize(Trim_String((char*)buf), Start_Gtk_Api_Config, handle, extra_handle, retstr);

		/* If logging is enabled, write returnstring to log */
		if (argc == 4 && log != NULL){
			fprintf(log, "SERVER: %s\n", retstr);
			fflush(log);
		}
		/* Add newline to returned string */
		strcat(retstr, "\n");

		/* Now send the result back to the socket */
		if ((numbytes=sendto(sockfd, retstr, strlen(retstr), 0, (struct sockaddr *)&their_addr, sizeof(struct sockaddr))) == -1) {
			Print_Error(2, "Could not send DATAGRAM packet", errno);
		}
	}
	else Print_Error(2, "Could not receive DATAGRAM packet", errno);

/* Endless while UDP */
}

/* End of the UDP protocol */
}

/* Arguments not recognized - return with error */
Print_Error(1, "Arguments to GTK-server not recognized!");
return -1;
}
/* END */
